#include <assert.h>
#include "types/dataset.h"
#include "types/aggregation.h"
#include "classnn.h"

void ClassNN::alloc_space (size_t n)
{
    wgt_n = n;
    if (n > 0) {
        comp_wgt = rm_alloc(n, output_dim);
        sample_wgt = rv_alloc(n);
    }
}

void ClassNN::free_space ()
{
    if (wgt_n > 0) {
        rm_free(comp_wgt, wgt_n);
        rv_free(sample_wgt);
    }
    wgt_n = 0;
}

ClassNN::ClassNN (size_t n_input, REAL **tr_i, REAL **tr_o, size_t trn,
                  REAL **val_i, REAL **val_o, size_t valn)
    : FeedForwardNN (n_input)
{
    tr_in = tr_i;
    tr_out = tr_o;
    n_tr = trn;
    val_in = val_i;
    val_out = val_o;
    n_val = valn;

    alloc_space(0);
}

ClassNN::ClassNN (ClassNN &nn)
    : FeedForwardNN(nn)
{
    tr_in  = nn.tr_in;  tr_out  = nn.tr_out;  n_tr  = nn.n_tr;
    val_in = nn.val_in; val_out = nn.val_out; n_val = nn.n_val;

    if (nn.wgt_n > 0) {
        alloc_space(nn.wgt_n);

        assert(nn.comp_wgt != NULL);
        for (size_t i = 0; i < wgt_n; i++)
            rv_copy(comp_wgt[i], nn.comp_wgt[i], output_dim);

        rv_copy(sample_wgt, nn.sample_wgt, wgt_n);
    }
    else wgt_n = 0;
}

ClassNN::~ClassNN ()
{
    free_space();
}

ClassNN* ClassNN::copy ()
{
    return new ClassNN(*this);
}

void ClassNN::set_weights (REAL **c_wgt, size_t n, REAL **y)
{
    free_space();
    if (n > 0) {
        alloc_space(n);

        for (size_t i = 0; i < n; i++) {
            REAL max, sum, *w, *yi;
            size_t j, k;

            max = sum = 0;
            w = c_wgt[i]; yi = y[i];
            k = 0;
            for (j = 0; j < output_dim; j++, w++) {
                if (*yi++ < 0.9) {
                    sum += *w;
                    if (max < *w) max = *w;
                }
                else
                    k = j;
            }
            sample_wgt[i] = sum;

            assert(max > 0);
            w = c_wgt[i]; REAL *cw = comp_wgt[i];
            for (j = 0; j < output_dim; j++, w++)
                if (j == k)
                    *cw++ = 3;
                else *cw++ = *w / max;
        }
    }
#ifndef NDEBUG
    REAL sum = 0;
    for (size_t i = 0; i < n; i++)
        sum += sample_wgt[i];
    printf("sum = %g\n", sum);
    assert(sum > 1-1e-6 && sum < 1+1e-6);
#endif
}

REAL ClassNN::_error (size_t k, REAL *y)
{
    REAL *out = output[n_layer];
    REAL err = 0;

    if (wgt_n > 0) {
        REAL *cw = comp_wgt[k];
        for (size_t i = 0; i < output_dim; i++) {
            REAL dif = (*out++) - (*y++);
            err += dif * dif * (*cw++);
        }
    }
    else {
        for (size_t i = 0; i < output_dim; i++) {
            REAL dif = *out - *y;
            if (*y > 0.9) err += 3 * dif * dif;
            else err += dif * dif;
        }
    }
    return err / 2;
}

void ClassNN::_error_deriv (size_t k, REAL *y, REAL *deriv)
{
    REAL *out = output[n_layer];
    if (wgt_n > 0) {
        REAL *cw = comp_wgt[k];
        for (size_t i = 0; i < layer_size[n_layer]; i++)
            *deriv++ = ((*out++) - (*y++)) * (*cw++);
    }
    else {
        for (size_t i = 0; i < layer_size[n_layer]; i++, out++, y++) {
            if (*y > 0.9)
                *deriv++ = 3 * (*out - *y);
            else *deriv++ = *out - *y;
        }
    }
}

REAL ClassNN::train (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    assert(n_layer > 0);
    assert(wgt == NULL || support_weighted_data());

#ifndef NDEBUG
    REAL wsum = 0;
    if (wgt != NULL) {
        for (size_t i = 0; i < n; i++)
            wsum += wgt[i];
        assert(wsum > 1-1e-5 && wsum < 1+1e-5);
    }
#endif
    REAL err;

    if (wgt_n > 0) {
        assert(wgt == NULL);
        wgt = sample_wgt;
    }
    switch (train_method) {
    case GRADIENT_DESCENT:
        err = gd_train(x, y, n, wgt);
        break;
    case LINE_SEARCH:
        err = line_train(x, y, n, wgt);
        break;
    case CONJUGATE_GRADIENT:
        err = cg(x, y, n, wgt);
        break;
    case ADAPTIVE_LEARNING_RATE:
        err = adapt_train(x, y, n, wgt);
        break;
    default:
        err = 0;
        assert(0);
    }

    return err;
}

#define _NN_FORWARD(j,x)           \
    output[0] = x;                 \
    for (j = 1; j <= n_layer; j++) \
        layer[j]->feed_forward(output[j-1], output[j])

#define GET_DERIVATIVE(i,j,wgt) \
    _NN_FORWARD(j, x[i]); \
    _error_deriv(i, y[i], deriv[n_layer]); \
    if (wgt != NULL) { \
        REAL w = wgt[i] * n; \
        for (j = 0; j < layer_size[n_layer]; j++) \
            deriv[n_layer][j] *= w; \
    } \
    for (j = n_layer; j > 0; j--) \
        layer[j]->back_propagate(output[j-1], deriv[j], deriv[j-1])

#define GET_ERROR(err,x,y,n,w) { \
    size_t i, j; \
    err = 0; \
    if (w == NULL) { \
        for (i = 0; i < n; i++) { \
            _NN_FORWARD(j, x[i]); \
            err += _error(i, y[i]); \
        } \
        err /= n; \
    } \
    else \
        for (i = 0; i < n; i++) { \
            _NN_FORWARD(j, x[i]); \
            err += _error(i, y[i]) * w[i]; \
        } \
}

REAL ClassNN::epoch (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    assert(layer_size != NULL);
    assert(x != NULL && y != NULL);
    assert(wgt == NULL || support_weighted_data());

    size_t i, j;
    for (i = 0; i < n; i++) {
        GET_DERIVATIVE(i, j, wgt);
        if (online_learn) {
            for (j = 1; j <= n_layer; j++)
                layer[j]->update_weight(learn_rate*100);
        }
    }
    if (!online_learn) {
        for (j = 1; j <= n_layer; j++)
            layer[j]->update_weight(learn_rate);
    }

    REAL err;
    GET_ERROR(err, x, y, n, wgt);
    return err;
}

REAL ClassNN::gd_train (REAL **org_x, REAL **org_y, size_t n, REAL *wgt)
{
    assert(wgt == NULL || support_weighted_data());

    REAL err = min_err + 1;
    size_t run = 0;

    //assert(wgt_n == n || wgt != NULL);

    REAL **x = NULL, **y = NULL;
    if (online_learn) {
        printf("Generating all random data ... "); fflush(stdout);
        x = (REAL**) malloc(n * max_run * sizeof(REAL*));
        y = (REAL**) malloc(n * max_run * sizeof(REAL*));
        Aggregation::resample(org_x, org_y, n, wgt, x, y, n * max_run);
        printf("done\n");
    }
    while (run < max_run && err >= min_err) {
        if (online_learn)
            err = epoch(&x[run*n], &y[run*n], n, NULL);
        else
            err = epoch(org_x, org_y, n, wgt);
        log_error(++run, err);
    }
    if (online_learn) {
        free(x); free(y);
    }
    return err;
}

#define SET_STEP(stp) { \
    size_t i; \
    for (i = 1; i <= n_layer; i++) \
        layer[i]->temp_update_weight(stp); \
}

REAL ClassNN::linesearch (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    size_t i;
    REAL stp1, stp2, stp3, stp4, stp5;
    REAL       err2, err3, err4, err5;
    REAL step = learn_rate;

    for (i = 1; i <= n_layer; i++)
        layer[i]->backup_weight();

    stp3 = 0;
    GET_ERROR(err3, x, y, n, wgt);
    stp5 = step;
    SET_STEP(stp5); GET_ERROR(err5, x, y, n, wgt);
    while (err5 > err3 && step > 1e-6) {
        putchar('-');
        step *= 0.5;
        stp5 = step;
        SET_STEP(stp5); GET_ERROR(err5, x, y, n, wgt);
    }

    if (err5 > err3) {
        fprintf(stderr, "\tWarning: err5 > err3\n");
    }
    do {
        putchar('*');
        step += step;
        stp1 = stp3;
        stp3 = stp5; err3 = err5;
        stp5 += step;
        SET_STEP(stp5); GET_ERROR(err5, x, y, n, wgt);
    } while (err5 < err3);

    while (stp3 > stp1*1.05 || stp5 > stp3*1.05) {
        putchar('.');
        stp2 = (stp1 + stp3) / 2;
        stp4 = (stp3 + stp5) / 2;
        SET_STEP(stp2); GET_ERROR(err2, x, y, n, wgt);
        SET_STEP(stp4); GET_ERROR(err4, x, y, n, wgt);

        if (err4 < err2 && err4 < err3) {
            stp1 = stp3;
            stp3 = stp4; err3 = err4;
        }
        else if (err2 < err3) {
            stp5 = stp3;
            stp3 = stp2; err3 = err2;
        }
        else {
            stp1 = stp2;
            stp5 = stp4;
        }
    }
    SET_STEP(0);
    for (i = 1; i <= n_layer; i++)
        layer[i]->update_weight(stp3);
    printf("\tstep = %g\n", stp3);
    return err3;
}

REAL ClassNN::line_train (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    REAL err = min_err + 1;
    size_t run = 0;

    assert(n_layer > 0);

    while (run < max_run && err >= min_err) {
        size_t i, j;

        for (i = 0; i < n; i++) {
            GET_DERIVATIVE(i, j, wgt);
        }

        err = linesearch(x, y, n, wgt);
        log_error(++run, err);
    }
    return err;
}

REAL ClassNN::cg (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    REAL err = min_err + 1;
    size_t run = 0;
    REAL g_norm_old = 0, g_norm, g_dot_old;

    assert(n_layer > 0);

    while (run < max_run && err >= min_err) {
        size_t i, j;

        /* get the gradient descent */
        for (i = 0; i < n; i++) {
            GET_DERIVATIVE(i, j, wgt);
        }

        /* get beta */
        g_norm = g_dot_old = 0;
        for (i = 1; i <= n_layer; i++) {
            g_norm += layer[i]->g_norm2();
            g_dot_old += layer[i]->g_dot_old();
        }

        REAL beta;
        if (run == 0)
            beta = 0;
        else {
            assert(g_norm_old > 0);
            beta = (g_norm - g_dot_old) / g_norm_old;
            //printf("beta%u = %g - %g = %g, ", count + 1,
            //       g_norm / g_norm_old, g_dot_old / g_norm_old, beta);
            if (beta < 0)
                beta = 0;
        }
        g_norm_old = g_norm;

        for (i = 1; i <= n_layer; i++)
            layer[i]->cg_update(beta);

        err = linesearch(x, y, n, wgt);
        log_error(++run, err);
    }
    return err;
}

REAL ClassNN::adapt_train (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    REAL err = min_err + 1, err_old = 1e9;
    size_t i = 0;

    assert(!online_learn);
    while (i < max_run && err >= min_err) {
        size_t j;
        for (j = 1; j <= n_layer; j++)
            layer[j]->backup_weight();

        err = epoch(x, y, n, wgt);
        log_error(++i, err);

        if (err < err_old) {
            learn_rate *= 1.15;
            if (learn_rate > 0.1)
                learn_rate = 0.1;
            err_old = err;
        }
        else {
            for (j = 1; j <= n_layer; j++)
                layer[j]->restore_weight();
            learn_rate *= 0.5;
        }
    }
    return err;
}

void ClassNN::log_error (size_t epoch, REAL err)
{
    size_t i, err_no = 0;
    REAL class_err, val_class_err;
    REAL *out = output[n_layer];

    if (logf == NULL && epoch % 20 != 1) return;

    for (i = 0; i < n_tr; i++) {
        size_t j, maxj;
        REAL max;

        _NN_FORWARD(j, tr_in[i]);
        max = out[0]; maxj = 0;
        for (j = 1; j < layer_size[n_layer]; j++) {
            if (max < out[j]) {
                max = out[j];
                maxj = j;
            }
        }
        if (tr_out[i][maxj] < 0) err_no++;
    }
    class_err = err_no / (REAL) n_tr;

    err_no = 0;
    for (i = 0; i < n_val; i++) {
        size_t j, maxj;
        REAL max;

        _NN_FORWARD(j, val_in[i]);
        max = out[0]; maxj = 0;
        for (j = 1; j < layer_size[n_layer]; j++) {
            if (max < out[j]) {
                max = out[j];
                maxj = j;
            }
        }
        if (val_out[i][maxj] < 0) err_no++;
    }
    val_class_err = err_no / (REAL) n_val;

    if (logf != NULL)
        fprintf(logf, "%d %g %g %g %g\n", epoch, learn_rate, err,
                class_err, val_class_err);

    if (epoch % 20 == 1)
        printf("epoch %d, error = %g (%4.1f%%,%4.1f%%)\n", epoch, err,
               class_err*100, val_class_err*100);
}
