#include <stdio.h>
#include <string.h>
#include "data.h"
#include "types/random.h"
#include "types/nnlayer.h"
#include "classnn.h"
#include "nnlayer_momentum.h"
#include "average.h"
#include "types/bagging.h"
#include "types/boosting.h"
#include "adaboost.h"
#include "adaboost_m2.h"

DataSet *pd, *va;

DataSet *read_data (FILE *f, size_t m, size_t n)
{
    static REAL mean[] = {
        4.0368, 7.0606, 5.1315, 5.3839, 3.5209, 6.9024, 7.5047, 4.6232,
        5.1661, 8.2830, 6.4708, 7.9230, 3.0628, 8.3466, 3.6966, 7.7961
    };
    static REAL stdd[] = {
        1.9231, 3.3000, 2.0155, 2.2582, 2.1960, 2.0215, 2.3201, 2.6970,
        2.3771, 2.4813, 2.6374, 2.0731, 2.3448, 1.5516, 2.5780, 1.6241
    };
    size_t i;
    RVEC *rm, *sam_in, *sam_out;

    assert(n == 17);

    rm = rm_alloc(m, n);
    if (rm_read(f, rm, m, n) != m) {
        fprintf(stderr, "read_data: Data file read error.\n");
        exit(-3);
    }

    DataSet *pd = new DataSet();
    sam_in = rm_alloc(m, 1);
    sam_out = rm_alloc(m, 26); /* 26 letters */
    for (i = 0; i < m; i++) {
        size_t j;

        for (j = 1; j < n; j++)
            //rm[i][j] = (rm[i][j] / 15 - 0.5) * 1.9; /* simple normalize */
            rm[i][j] = (rm[i][j] - mean[j-1]) / stdd[j-1];

        free(sam_in[i]);
        sam_in[i] = rm[i]+1;

        for (j = 0; j < 26; j++)
            sam_out[i][j] = -0.99;
        int cls = (int) (rm[i][0]-0.9);
        assert(cls+1-rm[i][0] < 1e-6 && cls+1-rm[i][0] > -1e-6);
        sam_out[i][cls] = 0.99;
    }

    pd->setData(m, n-1, 26, sam_in, sam_out);

    rm_free(rm, m);
    free(sam_in);
    rm_free(sam_out, m);

    return pd;
}

/* main() simply passes argc/argv to setup_data()
   to read/set up the data */
void setup_data (int argc, char *argv[])
{
    if (argc < 4) {
        fprintf(stderr, "Usage: %s file <# of data> <dim> logfile method resultfile\n", argv[0]);
        exit(-1);
    }

    FILE *f = fopen(argv[1], "r");
    if (f == NULL) {
        fprintf(stderr, "setup_data: data file open error.\n");
        exit(-2);
    }

    size_t m, vm, n;
    m = atoi(argv[2]);
    vm = atoi(argv[3]);
    n = 17;

    pd = read_data(f, m, n);
    va = read_data(f, vm, n);

    fclose(f);

    printf("Data loaded.\n");
}

REAL class_error (LearnModel *lm, DataSet *ds)
{
    size_t err = 0;
    size_t n = ds->getDataSize();
    size_t n_output = ds->getOutputDim();
    REAL **sam_in = ds->getInputBuf();
    REAL **sam_out = ds->getOutputBuf();
    REAL *out = rv_alloc(n_output);

    for (size_t i = 0; i < n; i++) {
        size_t j, maxj;
        REAL max;

        lm->get_output(sam_in[i], out);

        max = out[0]; maxj = 0;
        for (j = 1; j < n_output; j++) {
            if (max < out[j]) {
                max = out[j];
                maxj = j;
            }
        }
        if (sam_out[i][maxj] < 0) err++;
    }
    return (err / (REAL) n);
}

int main (int argc, char *argv[])
{
    size_t m, in, out;
    size_t n1neuron = 50, n2neuron;
    size_t n_agg = 50, max_run = 500;
    const char *nn_fn = "train_nn", *err_fn = "train_err";
    int method = 0, agg_method = 0;
    REAL learn_rate = 0.0001;
    const char *method_name[] = {
        "gradient descent",
        "adaptive grad-descent",
        "momentum",
        "line search",
        "CG",
        "stochastic GD"
    };
    const char *agg_method_name[] = {
        "no aggregation",
        "average",
        "bagging",
        "simple boosting",
        "AdaBoost.M1",
        "AdaBoost.M2"
    };

    set_seed(0);

    if (argc > 4) err_fn = argv[4];
    if (argc > 5) method = atoi(argv[5]);
    if (argc > 6) nn_fn = argv[6];
    if (argc > 7) n1neuron = atoi(argv[7]);
    if (argc > 8) agg_method = atoi(argv[8]);
    if (argc > 9) n_agg = atoi(argv[9]);
    if (argc > 10) learn_rate = atof(argv[10]);
    if (argc > 11) max_run = atoi(argv[11]);

    int test = (agg_method < 0);
    if (test) agg_method = -agg_method;

    setup_data(argc, argv);
    m   = pd->getDataSize();
    in  = pd->getInputDim();
    out = pd->getOutputDim();

    ClassNN nn(in, pd->getInputBuf(), pd->getOutputBuf(), m,
               va->getInputBuf(), va->getOutputBuf(), va->getDataSize());
    NNLayer *l1, *l2, *l3;

    n1neuron = 70; n2neuron = 50; //n1neuron;
    printf("Layer 1 neuron: %u\n", n1neuron);
    if (method == 2) /* momentum */ {
        l1 = new NNLayer_Momentum(in, n1neuron, 0.5);
        l2 = new NNLayer_Momentum(n1neuron, n2neuron, 0.5);
        l3 = new NNLayer_Momentum(n2neuron, out, 0.5);
    }
    else {
        l1 = new NNLayer(in, n1neuron);
        l2 = new NNLayer(n1neuron, n2neuron);
        l3 = new NNLayer(n2neuron, out);
    }
    l1->initialize(-1, 1);
    l2->initialize(-0.7, 0.7);
    l3->initialize(-0.5, 0.5);
    nn.add_layer(l1);
    nn.add_layer(l2);
    nn.add_layer(l3);
    delete(l1); delete(l2); delete(l3);

    if (!strcmp(err_fn, nn_fn)) {
        fprintf(stderr, "Error: log file and NN file are the same.\n");
        exit(-2);
    }
    FILE *f;
    if ((f = fopen(err_fn, "r")) != NULL) {
        fprintf(stderr, "Error: log file already exists.\n");
        exit(-3);
    }
    if ((f = fopen(err_fn, "w")) == NULL)
        fprintf(stderr, "Warning: log file open error.\n");
    nn.set_log_file(f);
    nn.set_batch_mode();

    switch (method) {
    case 0: nn.set_train_method(nn.GRADIENT_DESCENT); break;
    case 1: nn.set_train_method(nn.ADAPTIVE_LEARNING_RATE); break;
    case 2: nn.set_train_method(nn.GRADIENT_DESCENT); break;
    case 3: nn.set_train_method(nn.LINE_SEARCH); break;
    case 4: nn.set_train_method(nn.CONJUGATE_GRADIENT); break;
    case 5: nn.set_batch_mode(FALSE);
        nn.set_train_method(nn.GRADIENT_DESCENT); break;
    default:
        printf("No such learning method\n");
        exit(-1);
    }

    Aggregation *ag;
    switch (agg_method) {
    case 1: ag = new Average (16, 26, n_agg); break;
    case 2: ag = new Bagging (16, 26, n_agg); break;
    case 3: ag = new Boosting(16, 26, n_agg); break;
    case 4: ag = new AdaBoost(16, 26, n_agg); break;
    case 5: ag = new AdaBoost_M2(16, 26, n_agg); break;
    default: ag = NULL;
    }

    LearnModel *lm;
    if (agg_method == 0) {
        nn.set_parameter(learn_rate, 0, max_run);
        lm = &nn;
    }
    else if (ag == NULL) {
        printf("No such aggregation\n");
        exit(-2);
    }
    else {
        lm = ag;
        nn.set_parameter(learn_rate, 0, max_run);
        nn.set_log_file(NULL);
        ag->set_model_base(&nn);
    }

    printf("Use %s, with %s, learning rate %g, max run %u\n",
           method_name[method], agg_method_name[agg_method],
           learn_rate, max_run);
    lm->initialize();

    REAL err, cerr, verr;
    if (test) {
        FILE *nf;
        if ((nf = fopen(nn_fn, "r")) == NULL) {
            fprintf(stderr, "Warning: result file open error.\n");
            exit(-3);
        }
        else if (!lm->load_model(nf)) {
            fprintf(stderr, "Load result NN error.\n");
            exit(-4);
        }
        fclose(nf);
        fprintf(stderr, "Load NN OK.\n");
        //((Boosting*) ag)->convert_log_weight();
        err = -1;
    }
    else
        err = lm->train2(*pd);
    cerr = class_error(lm, pd) * 100;
    verr = class_error(lm, va) * 100;
    printf("\n\naverage error = %g (%g%%, %g%%)\n\n", err, cerr, verr);

    if (ag != NULL) { /* verify the model */
        size_t n = ag->get_n_model(), i;
        for (i = 1; i <= n; i++) {
            ag->set_aggregation_size(i);
            printf("verifying aggregation size %2u ... ", i);
            fflush(stdout);
            fprintf(f, "%u %g %g\n", i,
                    class_error(ag, pd), class_error(ag, va));
            printf("done\n");
        }
        printf("verifying single models ");
        fflush(stdout);

        cerr = verr = 0;
        for (i = 1; i <= n; i++) {
            LearnModel *lm = ag->get_model(i-1);
            cerr += class_error(lm, pd);
            verr += class_error(lm, va);
            printf("."); fflush(stdout);
            fprintf(f, "%u %g %g\n", i, cerr/i, verr/i);
        }
        printf("done\n Avg Cls err = %g%%, Avg Val err = %g%%\n",
               cerr/n*100, verr/n*100);
    }

    if (f != NULL) fclose(f);

    if (!test) {
        if ((f = fopen(nn_fn, "w")) == NULL)
            fprintf(stderr, "Warning: result file open error.\n");
        else if (!lm->save_model(f))
            fprintf(stderr, "Save result NN error.\n");
        if (f != NULL) fclose(f);
    }

    delete(pd); delete(va);
    if (ag != NULL) delete ag;
    return 0;
}
