#include <assert.h>
#include <stdio.h>
#include "random.h"
#include "dataset.h"
#include "aggregation.h"

void Aggregation::alloc_space (size_t max)
{
    max_model = max;
    if (max_model > 0) {
        lm = new (LearnModel*)[max_model];
#ifndef NDEBUG
        for (size_t i = 0; i < max_model; i++)
            lm[i] = NULL;
#endif
        model_out = rm_alloc(max_model, output_dim);
    }
    n_model = n_in_agg = 0;
}

void Aggregation::free_space ()
{
    if (max_model > 0) {
        for (size_t i = 0; i < n_model; i++) {
            assert(lm[i] != NULL);
            delete lm[i];
        }
        delete lm;
        rm_free(model_out, max_model);
        max_model = n_model = n_in_agg = 0;
    }
    assert(n_model == 0);
    /* should not delete lm_base here, since load_model need it */
}

Aggregation::Aggregation (size_t n_input, size_t n_output, size_t max)
    : LearnModel(n_input, n_output)
{
    alloc_space(max);
    lm_base = NULL;
}

Aggregation::Aggregation (Aggregation &ag)
    : LearnModel(ag)
{
    alloc_space(ag.max_model);
    lm_base = NULL;
    n_model = ag.n_model;
    n_in_agg = ag.n_in_agg;
    for (size_t i = 0; i < n_model; i++)
        lm[i] = ag.lm[i]->copy();
    set_model_base(ag.lm_base);
}

Aggregation::~Aggregation ()
{
    free_space();
    if (lm_base != NULL) delete(lm_base);
}

void Aggregation::set_aggregation_size (size_t n)
{
    if (n > 0 && n <= n_model)
        n_in_agg = n;
}

LearnModel* Aggregation::get_model (size_t n)
{
    if (n >= n_model)
        return NULL;
    return lm[n];
}

void Aggregation::get_output (REAL *x, REAL *y)
{
    assert(n_in_agg > 0 && n_in_agg <= n_model);
    for (size_t i = 0; i < n_in_agg; i++) {
        assert(lm[i] != NULL);
        lm[i]->get_output(x, model_out[i]);
    }
    aggregate(y);
}

void Aggregation::set_model_base (LearnModel *lmb)
{
    if (lm_base != NULL) {
        delete(lm_base);
        lm_base = NULL;
    }
    if (lmb != NULL) {
        assert(lmb->get_n_input() == input_dim
               && lmb->get_n_output() == output_dim);
        lm_base = lmb->copy();
    }
}

BOOL Aggregation::save_model (FILE *f)
{
    assert(f != NULL);

    fprintf(f, "# %s\n", version());
    fprintf(f, "%u %u %u %u\n", input_dim, output_dim, n_model,
            lm_base != NULL);
    if (lm_base != NULL) {
        fprintf(f, "# current base\n");
        if (!lm_base->save_model(f))
            return FALSE;
    }
    for (size_t i = 0; i < n_model; i++) {
        if (!lm[i]->save_model(f))
            return FALSE;
    }
    return TRUE;
}

BOOL Aggregation::load_model (FILE *f)
{
    /* can't do without a base model, since we can not
     * create an unknown (to LearnModel) model from the file.
     */
    if (lm_base == NULL)
        return FALSE;

    assert(f != NULL);
    ignore_comment(f);

    size_t t1, t2, t3, t4;
    if (fscanf(f, "%u %u %u %u", &t1, &t2, &t3, &t4) != 4) return FALSE;
    if (t1 <= 0 || t2 <= 0 || t3 <= 0) return FALSE;
    if (t4 != 0 && t4 != 1) return FALSE;

    if (t4) {
        ignore_comment(f);
        if (!lm_base->load_model(f))
            return FALSE;
        if (lm_base->get_n_input() != t1 || lm_base->get_n_output() != t2)
            return FALSE;
    }

    free_space();
    input_dim = t1; output_dim = t2;
    alloc_space(t3);

    assert(n_model == 0);
    while (n_model < max_model) {
        LearnModel *p = lm_base->copy();
        if (!p->load_model(f))
            return FALSE;

        lm[n_model] = p;
        n_model++;

        if (p->get_n_input() != t1 || p->get_n_output() != t2)
            return FALSE;
    }
    n_in_agg = n_model;

    if (!t4) set_model_base(NULL);

    return TRUE;
}

void Aggregation::resample (REAL **x, REAL **y, size_t n, REAL *wgt,
                            REAL **rx, REAL **ry, size_t rn)
{
    assert(n > 0);
    size_t i, sel;
    if (wgt == NULL) {
        for (i = 0; i < rn; i++) {
            sel = (size_t) (randu() * n);
            assert(sel < n);
            rx[i] = x[sel];
            ry[i] = y[sel];
        }
    }
    else {
        REAL *cdf = rv_alloc(n+1);
        cdf[0] = 0;
        for (i = 0; i < n; i++)
            cdf[i+1] = cdf[i] + wgt[i];
        assert(cdf[n] > 1-1e-6 && cdf[n] < 1+1e-6);

        for (i = 0; i < rn; i++) {
            REAL r = randu();

            size_t b = 0, e = n, m;
            while (b+1 < e) {
                m = (b + e) / 2;
                if (r < cdf[m]) e = m;
                else b = m;
            }

            rx[i] = x[b];
            ry[i] = y[b];
        }
    }
}
