#ifndef __AGGREGATION_H__
#define __AGGREGATION_H__
#define __LEARNING_MODEL_AGGREGATION_H__

#include <stdio.h>
#include "learnmodel.h"

/* The designed calling order of functions
 *
 *  1. Aggregation *ag = new SomeAggMethod (6, 5, 10);
 *         // allocate room for outputs, and learning model pointers.
 *
 *  2. ag->set_model_base(pointer_to_a_neural_net);
 *         // specify what learn model to use. If the training of those
 *         // those learning models are parallel, we can allocate room
 *         // for those models here, and initialize them in step 3;
 *         // otherwise the allocation can be carried out later.
 *
 *  3. ag->initialize();
 *         // prepare a training
 *
 *  4. avg_err = ag->train(sam_in, sam_out, n_sample);
 *         // usually, the return value is the average error
 *
 *  5. ag->get_output(x, y)
 *         // apply the aggregated model to new data.
 */

class Aggregation : public LearnModel
{
 private:
    static const char name_version[] = "Aggregation 10/27/2001";

 protected:
    LearnModel **lm;      /* pointer to models */
    LearnModel *lm_base;  /* every model should be a copy of lm_base */
    size_t n_model;        /* number of models */
    size_t n_in_agg;       /* # of models in aggregation */
    size_t max_model;      /* maximum number of models */

    REAL **model_out;

 private:
    void alloc_space (size_t max);
    void free_space ();

 public:
    Aggregation (size_t n_input, size_t n_output, size_t max);
    Aggregation (Aggregation &ag);
    virtual ~Aggregation ();
    virtual const char* version () { return name_version; }

    virtual BOOL save_model (FILE *f);
    virtual BOOL load_model (FILE *f);
    virtual Aggregation* copy () = 0;

    void set_model_base (LearnModel *lmb);
    void set_aggregation_size (size_t n);
    size_t get_n_model () { return n_model; }
    size_t get_aggregation_size () { return n_in_agg; }

    LearnModel* get_model (size_t n);  /* not copy */

    /* commonly used sub-routines */
    static void resample (REAL **x, REAL **y, size_t n, REAL *wgt,
                   REAL **rx, REAL **ry, size_t rn);

 protected:
    /* convert model_out to aggragated one */
    virtual void aggregate (REAL *y) = 0;

 public:
    virtual void initialize () = 0;
    virtual REAL train (REAL **x, REAL **y, size_t n, REAL *wgt = NULL) = 0;
    virtual void get_output (REAL *x, REAL *y);
};

#else /* def __AGGREGATION_H__ */

#ifndef __LEARNING_MODEL_AGGREGATION_H__
#error This header file conflicts with another "aggregation.h" file.
#endif

#endif
