#include <assert.h>
#include <math.h>
#include <stdio.h>
#include "dataset.h"
#include "boosting.h"

Boosting::Boosting (size_t n_input, size_t n_output, size_t max)
    : Aggregation(n_input, n_output, max)
{
    assert(max == max_model);
    lm_wgt = rv_alloc(max_model);
    rv_clear(lm_wgt, max_model);
}

Boosting::Boosting (Boosting &ag)
    : Aggregation(ag)
{
    assert(max_model == ag.max_model);
    lm_wgt = rv_alloc(max_model);
    assert(n_model == ag.n_model);
    rv_copy(lm_wgt, ag.lm_wgt, n_model);
}

Boosting* Boosting::copy()
{
    return new Boosting(*this);
}

BOOL Boosting::save_model (FILE *f)
{
    assert(f != NULL);

    if (!Aggregation::save_model(f))
        return FALSE;
    for (size_t i = 0; i < n_model; i++)
        fprintf(f, "%g ", lm_wgt[i]);
    fprintf(f, "\n");

    return TRUE;
}

BOOL Boosting::load_model (FILE *f)
{
    if (!Aggregation::load_model (f))
        return FALSE;

    rv_free(lm_wgt);
    lm_wgt = rv_alloc(max_model);
    rv_clear(lm_wgt, max_model);
    for (size_t i = 0; i < n_model; i++)
        if (fscanf(f, "%lg", &lm_wgt[i]) != 1)
            return FALSE;

    return TRUE;
}

void Boosting::initialize ()
{
    assert(lm_base != NULL);
    lm_base->initialize();

    for (size_t i = 0; i < n_model; i++) {
        assert(lm[i] != NULL);
        delete(lm[i]);
    }
    n_in_agg = n_model = 0;
    rv_clear(lm_wgt, max_model);
}

void Boosting::aggregate (REAL *y)
{
    assert(n_in_agg > 0 && n_in_agg <= n_model);

    size_t i, j;
    REAL wsum = 0;
    for (j = 0; j < n_in_agg; j++)
        wsum += lm_wgt[j];
    assert(wsum != 0);

    for (i = 0; i < output_dim; i++) {
        y[i] = 0;
        for (j = 0; j < n_in_agg; j++)
           y[i] += model_out[j][i] * lm_wgt[j];
        y[i] /= wsum;
    }
}

REAL Boosting::train (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    assert(n_in_agg == 0 && n_model == 0);
    assert(wgt == NULL || support_weighted_data());
    assert(lm_base != NULL); /* we need lm_base to create new hypotheses */

    size_t i;
    boost_wgt = rv_alloc(n);
    if (wgt == NULL) {
        for (i = 0; i < n; i++)
            boost_wgt[i] = 1.0 / n;
    }
    else
        rv_copy(boost_wgt, wgt, n);

    BOOL wgted_data = lm_base->support_weighted_data();
    REAL **boost_x = NULL, **boost_y = NULL;
    if (!wgted_data) {
        boost_x = (REAL**) malloc(n * sizeof(REAL*));
        boost_y = (REAL**) malloc(n * sizeof(REAL*));
    }

    REAL err_sum = 0, wsum = 0;
    for (i = 0; i < max_model; i++) {
        lm[i] = lm_base->copy();
        n_model++;
        printf("=== %s #%u / %u ===\n", version(), n_model, max_model);
        assert(lm[i] != NULL);

        REAL err;
        if (!wgted_data) {
            printf("generating data ... "); fflush(stdout);
            resample(x, y, n, boost_wgt, boost_x, boost_y, n);
            printf("done.\n");
            err = lm[i]->train(boost_x, boost_y, n, NULL);
        }
        else err = lm[i]->train(x, y, n, boost_wgt);

        /* update boost_wgt, set up hypothesis wgt (lm_wgt) */
        update_weights(i, x, y, n);
        n_in_agg++;

        err_sum += err * lm_wgt[i];
        wsum += lm_wgt[i];
    }

    if (!wgted_data) {
        free(boost_x);
        free(boost_y);
    }
    rv_free(boost_wgt);
    assert(wsum != 0);
    return (err_sum / wsum);
}

void Boosting::update_weights (size_t k, REAL **x, REAL **y, size_t n)
{
    /* a simple version */
    LearnModel *l = lm[k];
    size_t i;

    assert(output_dim == l->get_n_output());
    REAL *out = rv_alloc(output_dim);
    REAL bw_sum = 0, err_sum = 0;
    for (i = 0; i < n; i++) {
        l->get_output(x[i], out);
        REAL err = l->error(out, y[i]);
        err_sum += err * boost_wgt[i];

        boost_wgt[i] *= err;
        bw_sum += boost_wgt[i];
    }
    rv_free(out);

    lm_wgt[k] = 1 / log(1.01+err_sum);
    assert(bw_sum != 0);
    for (i = 0; i < n; i++)
        boost_wgt[i] /= bw_sum;
}

void Boosting::convert_log_weight ()
{
    for (size_t i = 0; i < n_model; i++) {
        assert(lm_wgt[i] > 0 && lm_wgt[i] < 1 / log(1.01));
        lm_wgt[i] = 1 / (exp(1 / lm_wgt[i]) - 1.01);
    }
}
