#include <assert.h>
#include <stdlib.h>
#include "vecmat.h"
#include "dataset.h"

DEFINE_ALLOC_FREE_FUNC(REAL, r);

Matrix::Matrix ()
{
    init();
}

Matrix::~Matrix ()
{
    freeBuffer();
}

void Matrix::init ()
{
    size = alloc_size = 0;
    dim = 0;
    x = NULL;
}

void Matrix::freeBuffer ()
{
    if (x != NULL) {
        assert(size > 0);
        rm_free(x, size);
    }
    init();
}

BOOL Matrix::setData (size_t len, size_t dimen, REAL **buf)
{
    freeBuffer();
    return appendData(len, dimen, buf);
}

static REAL ** newVectors (REAL **pMat, size_t n, size_t len)
{
    assert(pMat != NULL || n == 0);
    for (size_t i = 0; i < n; i++) {
        pMat[i] = rv_alloc(len);
        assert(pMat[i] != NULL);
    }
    return pMat;
}

BOOL Matrix::expandSize (size_t len, size_t dimen)
{
    if (size == 0)
        dim = dimen;
    assert(dim == dimen);

#ifndef NDEBUG
    REAL *xf = NULL, *xl = NULL; /* verify x_first, x_last */
    if (size > 0) {
        xf = x[0];
        xl = x[size-1];
    }
#endif

    if (size + len > alloc_size) {
        size_t next_size = size + len + 10;
        void *tmp = realloc(x, next_size * sizeof(REAL*));
        if (tmp == NULL)
            return FALSE;   /* failed to append the data */
        x = (REAL**) tmp;

        alloc_size = next_size;
    }

#ifndef NDEBUG
    if (size > 0) {
        assert(xf == x[0]);
        assert(xl == x[size-1]);
    }
#endif
    return TRUE;
}

BOOL Matrix::appendData (size_t len, size_t dimen, REAL **buf)
{
    if (!expandSize(len, dimen))
        return FALSE;

    if (newVectors(x+size, len, dimen) == NULL)
        return FALSE;

    size_t i, j;
    for (i = 0; i < len; i++, size++) {
        for (j = 0; j < dimen; j++)
            x[size][j] = buf[i][j];
    }

    return TRUE;
}

//==???==
DataSet::DataSet()
{
    init();
}

DataSet::~DataSet()
{
    freeAll();
}

void DataSet::init ()
{
    x = new Matrix();
    y = new Matrix();
    assert(x != NULL && y != NULL);
}

void DataSet::freeAll ()
{
    assert((x == NULL && y == NULL)
        || (x != NULL && y != NULL));
    if (x != NULL)
        delete x;
    if (y != NULL)
        delete y;
}

BOOL DataSet::setData (size_t len, size_t i, size_t o, REAL **x, REAL **y)
{
    freeAll(); init();
    return appendData(len, i, o, x, y);
}

BOOL DataSet::appendData (size_t len, size_t i, size_t o, REAL **xb, REAL **yb)
{
    assert(x != NULL && y != NULL);
    assert(xb != NULL && yb != NULL);
#ifndef NDEBUG
    if (x->getSize() > 0)
        assert(x->getDimension() == i && y->getDimension() == o);
#endif

    return (x->appendData(len, i, xb) && y->appendData(len, o, yb));
}

size_t DataSet::readDataFile (size_t len, size_t in, size_t out, FILE *f)
{
    REAL *r_in, *r_out;

    assert(f != NULL);
    assert(x != NULL && y != NULL);

    r_in  = rv_alloc(in);
    r_out = rv_alloc(out);

    size_t i, j;
    for (i = 0; i < len; i++) {
        for (j = 0; j < in; j++) {
            int c = fscanf(f, "%lf", &r_in[j]);
            if (c < 1) {
                /*fprintf(stderr, "rv_read: line too short (%d)\n", i);*/
                goto exit_readfile;
            }
        }

        for (j = 0; j < out; j++) {
            int c = fscanf(f, "%lf", &r_out[j]);
            if (c < 1) goto exit_readfile;
        }

        if (!x->appendData(1, in, &r_in) || !y->appendData(1, out, &r_out))
            break;
    }

exit_readfile:
    rv_free(r_in);
    rv_free(r_out);
    return i;
}

BOOL DataSet::combineData (Matrix &in, Matrix &out)
{
    return FALSE;
}
