#include <assert.h>
#include "dataset.h"
#include "feedforwardnn.h"

void FeedForwardNN::alloc_space (size_t n_input)
{
    n_layer = 0;
    layer_size[0] = n_input;
    layer[0] = NULL;
    output[0] = NULL;
    deriv[0] = rv_alloc(n_input);
}

void FeedForwardNN::free_space ()
{
    for (size_t i = 1; i <= n_layer; i++) {
        delete(layer[i]);
        rv_free(output[i]);
        rv_free(deriv[i]);
    }
    rv_free(deriv[0]);
    n_layer = 0; output_dim = input_dim;
}

FeedForwardNN::FeedForwardNN (size_t n_input)
    : LearnModel(n_input, n_input)
{
    alloc_space(n_input);
    online_learn = TRUE;
    train_method = GRADIENT_DESCENT;
}

FeedForwardNN::FeedForwardNN (FeedForwardNN &nn)
    : LearnModel(nn)
{
    assert(input_dim == nn.input_dim);  /* LearnModel should do this copy */
    alloc_space(input_dim);

    online_learn = nn.online_learn;
    learn_rate = nn.learn_rate;
    min_err = nn.min_err;
    max_run = nn.max_run;
    train_method = nn.train_method;

    for (size_t i = 1; i <= nn.n_layer; i++)
        add_layer(nn.layer[i]);
}

FeedForwardNN::~FeedForwardNN ()
{
    free_space();
}

FeedForwardNN* FeedForwardNN::copy ()
{
    return new FeedForwardNN(*this);
}

void FeedForwardNN::add_layer (NNLayer *nl)
{
    assert(layer_size[n_layer] == nl->get_n_input());

    n_layer++;
    layer[n_layer] = nl->copy();
    output_dim = layer_size[n_layer] = nl->get_n_neuron();
    output[n_layer] = rv_alloc(layer_size[n_layer]);
    deriv[n_layer] = rv_alloc(layer_size[n_layer]);
}

BOOL FeedForwardNN::save_model (FILE *f)
{
    size_t i;
    assert(f != NULL);

    fprintf(f, "# %s\n%u\n", version(), n_layer);
    for (i = 0; i <= n_layer; i++)
        fprintf(f, "%d ", layer_size[i]);
    fprintf(f, "\n%u %g %g %u\n", online_learn, learn_rate, min_err, max_run);
    for (i = 1; i <= n_layer; i++) {
        if (!layer[i]->save_model(f))
            return FALSE;
    }
    return TRUE;
}

BOOL FeedForwardNN::load_model (FILE *f)
{
    assert(f != NULL);

    ignore_comment(f);

    size_t tmp_layer;
    if (fscanf(f, "%u", &tmp_layer) != 1) return FALSE;
    if (tmp_layer <= 0) return FALSE;

    /* free the old model !!! */
    free_space();

    size_t i;
    for (i = 0; i <= tmp_layer; i++) {
        if (fscanf(f, "%u", &layer_size[i]) != 1) return FALSE;
        if (layer_size[i] <= 0) return FALSE;
    }

    alloc_space(input_dim = layer_size[0]);

    unsigned int online;
    if (fscanf(f, "%u %lf %lf %u",
               &online, &learn_rate, &min_err, &max_run) != 4)
        return FALSE;

    if (learn_rate <= 0 || min_err < 0 || max_run < 1)
        return FALSE;
    online_learn = (online != 0);

    assert(n_layer == 0);
    for (i = 1; i <= tmp_layer; i++) {
        NNLayer l(1,1);

        if (!l.load_model(f))
            return FALSE;
        if (l.get_n_input() != layer_size[i-1]
            || l.get_n_neuron() != layer_size[i])
            return FALSE;
        add_layer(&l);
    }
    return TRUE;
}

REAL FeedForwardNN::error (REAL *out, REAL *y)
{
    return LearnModel::error(out, y) / 2;
}

REAL FeedForwardNN::_error (REAL *y)
{
    return error(output[n_layer], y);
}

void FeedForwardNN::error_deriv (REAL *y, REAL *deriv)
{
    size_t i;
    REAL *out = output[n_layer];
    for (i = 0; i < layer_size[n_layer]; i++) {
        *deriv++ = (*out++) - (*y++);
    }
}

void FeedForwardNN::initialize ()
{
    for (size_t i = 1; i <= n_layer; i++) {
        assert(layer[i] != NULL);
        layer[i]->initialize();
    }
}

void FeedForwardNN::set_parameter (REAL lr, REAL minerr, size_t maxrun)
{
    learn_rate = lr;
    min_err = minerr;
    max_run = maxrun;
}

void FeedForwardNN::log_error (size_t epoch, REAL err)
{
    if (logf != NULL)
        fprintf(logf, "%d %g %g\n", epoch, learn_rate, err);

    if (epoch % 20 == 1)
        printf("epoch %d, error = %g\n", epoch, err);
}

REAL FeedForwardNN::train (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    assert(n_layer > 0);
    assert(wgt == NULL || support_weighted_data());

#ifndef NDEBUG
    REAL wsum = 0;
    if (wgt != NULL) {
        for (size_t i = 0; i < n; i++)
            wsum += wgt[i];
        assert(wsum > 1-1e-5 && wsum < 1+1e-5);
    }
#endif

    REAL err;
    switch (train_method) {
    case GRADIENT_DESCENT:
        err = gd_train(x, y, n, wgt);
        break;
    case LINE_SEARCH:
        err = line_train(x, y, n, wgt);
        break;
    case CONJUGATE_GRADIENT:
        err = cg(x, y, n, wgt);
        break;
    case ADAPTIVE_LEARNING_RATE:
        err = adapt_train(x, y, n, wgt);
        break;
    default:
        err = 0;
        assert(0);
    }

    return err;
}

#define _NN_FORWARD(j,x)           \
    output[0] = x;                 \
    for (j = 1; j <= n_layer; j++) \
        layer[j]->feed_forward(output[j-1], output[j])

#define GET_DERIVATIVE(i,j,wgt) \
    _NN_FORWARD(j, x[i]); \
    error_deriv(y[i], deriv[n_layer]); \
    if (wgt != NULL) { \
        REAL w = wgt[i] * n; \
        for (j = 0; j < layer_size[n_layer]; j++) \
            deriv[n_layer][j] *= w; \
    } \
    for (j = n_layer; j > 0; j--) \
        layer[j]->back_propagate(output[j-1], deriv[j], deriv[j-1])

#define GET_ERROR(err,x,y,n,w) { \
    size_t i, j; \
    err = 0; \
    if (w == NULL) { \
        for (i = 0; i < n; i++) { \
            _NN_FORWARD(j, x[i]); \
            err += _error(y[i]); \
        } \
        err /= n; \
    } \
    else \
        for (i = 0; i < n; i++) { \
            _NN_FORWARD(j, x[i]); \
            err += _error(y[i]) * w[i]; \
        } \
}

REAL FeedForwardNN::epoch (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    assert(layer_size != NULL);
    assert(x != NULL && y != NULL);
    assert(wgt == NULL || support_weighted_data());

    size_t i, j;
    for (i = 0; i < n; i++) {
        GET_DERIVATIVE(i, j, wgt);
        if (online_learn) {
            for (j = 1; j <= n_layer; j++)
                layer[j]->update_weight(learn_rate);
        }
    }
    if (!online_learn) {
        for (j = 1; j <= n_layer; j++)
            layer[j]->update_weight(learn_rate);
    }

    REAL err;
    GET_ERROR(err, x, y, n, wgt);
    return err;
}

REAL FeedForwardNN::gd_train (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    assert(wgt == NULL || support_weighted_data());

    REAL err = min_err + 1;
    size_t run = 0;

    while (run < max_run && err >= min_err) {
        err = epoch(x, y, n, wgt);
        log_error(++run, err);
    }
    return err;
}

#define SET_STEP(stp) { \
    size_t i; \
    for (i = 1; i <= n_layer; i++) \
        layer[i]->temp_update_weight(stp); \
}

REAL FeedForwardNN::linesearch (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    size_t i;
    REAL stp1, stp2, stp3, stp4, stp5;
    REAL       err2, err3, err4, err5;
    REAL step = learn_rate;

    for (i = 1; i <= n_layer; i++)
        layer[i]->backup_weight();

    stp3 = 0;
    GET_ERROR(err3, x, y, n, wgt);
    stp5 = step;
    SET_STEP(stp5); GET_ERROR(err5, x, y, n, wgt);
    while (err5 > err3 && step > 1e-6) {
        putchar('-');
        step *= 0.5;
        stp5 = step;
        SET_STEP(stp5); GET_ERROR(err5, x, y, n, wgt);
    }

    if (err5 > err3) {
        fprintf(stderr, "\tWarning: err5 > err3\n");
    }
    do {
        putchar('*');
        step += step;
        stp1 = stp3;
        stp3 = stp5; err3 = err5;
        stp5 += step;
        SET_STEP(stp5); GET_ERROR(err5, x, y, n, wgt);
    } while (err5 < err3);

    while (stp3 > stp1*1.05 || stp5 > stp3*1.05) {
        putchar('.');
        stp2 = (stp1 + stp3) / 2;
        stp4 = (stp3 + stp5) / 2;
        SET_STEP(stp2); GET_ERROR(err2, x, y, n, wgt);
        SET_STEP(stp4); GET_ERROR(err4, x, y, n, wgt);

        if (err4 < err2 && err4 < err3) {
            stp1 = stp3;
            stp3 = stp4; err3 = err4;
        }
        else if (err2 < err3) {
            stp5 = stp3;
            stp3 = stp2; err3 = err2;
        }
        else {
            stp1 = stp2;
            stp5 = stp4;
        }
    }
    SET_STEP(0);
    for (i = 1; i <= n_layer; i++)
        layer[i]->update_weight(stp3);
    printf("\tstep = %g\n", stp3);
    return err3;
}

REAL FeedForwardNN::line_train (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    REAL err = min_err + 1;
    size_t run = 0;

    assert(n_layer > 0);

    while (run < max_run && err >= min_err) {
        size_t i, j;

        for (i = 0; i < n; i++) {
            GET_DERIVATIVE(i, j, wgt);
        }

        err = linesearch(x, y, n, wgt);
        log_error(++run, err);
    }
    return err;
}

REAL FeedForwardNN::cg (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    REAL err = min_err + 1;
    size_t run = 0;
    REAL g_norm_old = 0, g_norm, g_dot_old;

    assert(n_layer > 0);

    while (run < max_run && err >= min_err) {
        size_t i, j;

        /* get the gradient descent */
        for (i = 0; i < n; i++) {
            GET_DERIVATIVE(i, j, wgt);
        }

        /* get beta */
        g_norm = g_dot_old = 0;
        for (i = 1; i <= n_layer; i++) {
            g_norm += layer[i]->g_norm2();
            g_dot_old += layer[i]->g_dot_old();
        }

        REAL beta;
        if (run == 0)
            beta = 0;
        else {
            assert(g_norm_old > 0);
            beta = (g_norm - g_dot_old) / g_norm_old;
            //printf("beta%u = %g - %g = %g, ", count + 1,
            //       g_norm / g_norm_old, g_dot_old / g_norm_old, beta);
            if (beta < 0)
                beta = 0;
        }
        g_norm_old = g_norm;

        for (i = 1; i <= n_layer; i++)
            layer[i]->cg_update(beta);

        err = linesearch(x, y, n, wgt);
        log_error(++run, err);
    }
    return err;
}

REAL FeedForwardNN::adapt_train (REAL **x, REAL **y, size_t n, REAL *wgt)
{
    REAL err = min_err + 1, err_old = 1e9;
    size_t i = 0;

    assert(!online_learn);
    while (i < max_run && err >= min_err) {
        size_t j;
        for (j = 1; j <= n_layer; j++)
            layer[j]->backup_weight();

        err = epoch(x, y, n, wgt);
        log_error(++i, err);

        if (err < err_old) {
            learn_rate *= 1.15;
            if (learn_rate > 0.1)
                learn_rate = 0.1;
            err_old = err;
        }
        else {
            for (j = 1; j <= n_layer; j++)
                layer[j]->restore_weight();
            learn_rate *= 0.5;
        }
    }
    return err;
}

void FeedForwardNN::get_output (REAL *x, REAL *y)
{
    assert(x != NULL && y != NULL);

    REAL *tmp = output[n_layer];
    size_t j;

    output[n_layer] = y;
    _NN_FORWARD(j, x);
    output[n_layer] = tmp;
}
