#include <assert.h>
#include <math.h>
#include "random.h"
#include "quickfun.h"
#include "dataset.h"
#include "nnlayer.h"

void NNLayer::alloc_space ()
{
    assert(n_neuron > 0 && n_input > 0);
    weight   = rm_alloc(n_neuron, n_input+1);
    wbackup  = rm_alloc(n_neuron, n_input+1);
    wdelta   = rm_alloc(n_neuron, n_input+1);
    wdel_old = rm_alloc(n_neuron, n_input+1);
    wdir     = rm_alloc(n_neuron, n_input+1);
    sig_der  = rv_alloc(n_neuron);
    for (size_t i = 0; i < n_neuron; i++)
        rv_clear(wdelta[i], n_input+1);
}

void NNLayer::free_space ()
{
    rm_free(weight, n_neuron);
    rm_free(wbackup, n_neuron);
    rm_free(wdelta, n_neuron);
    rm_free(wdel_old, n_neuron);
    rm_free(wdir, n_neuron);
    rv_free(sig_der);
    n_input = n_neuron = 0;
}

NNLayer::NNLayer (size_t num_input, size_t num_neuron)
{
    n_neuron = num_neuron;
    n_input  = num_input;
    alloc_space();

    winit_min = -1;
    winit_max = 1;

    quick_tanh_setup();
}

NNLayer::NNLayer (NNLayer &l) : Object(l)
{
    n_neuron  = l.n_neuron;
    n_input   = l.n_input;
    alloc_space();
    for (size_t i = 0; i < n_neuron; i++) {
        rv_copy(weight[i],   l.weight[i],   n_input+1);
        rv_copy(wbackup[i],  l.wbackup[i],  n_input+1);
        rv_copy(wdelta[i],   l.wdelta[i],   n_input+1);
        rv_copy(wdir[i],     l.wdir[i],     n_input+1);
        rv_copy(wdel_old[i], l.wdel_old[i], n_input+1);
    }
    rv_copy(sig_der, l.sig_der, n_neuron);

    winit_min = l.winit_min;
    winit_max = l.winit_max;
}

NNLayer::~NNLayer ()
{
    free_space();
}

NNLayer* NNLayer::copy ()
{
    return new NNLayer(*this);
}

BOOL NNLayer::save_model (FILE *f)
{
    assert(f != NULL);
    fprintf(f, "# %s\n", version());
    fprintf(f, "%u %u %g %g\n", n_input, n_neuron, winit_min, winit_max);
    size_t i, j;
    for (i = 0; i < n_neuron; i++) {
        for (j = 0; j <= n_input; j++)
            fprintf(f, "%g ", weight[i][j]);
        fprintf(f, "\n");
    }
    return TRUE;
}

BOOL NNLayer::load_model (FILE *f)
{
    assert(f != NULL);
    free_space();

    ignore_comment(f);
    if (fscanf(f, "%u %u %lf %lf\n",
               &n_input, &n_neuron, &winit_min, &winit_max) != 4)
        return FALSE;
    if (winit_min >= winit_max) return FALSE;

    if (n_input <= 0 || n_neuron <= 0)
        return FALSE;

    alloc_space();

    size_t i, j;
    for (i = 0; i < n_neuron; i++) {
        for (j = 0; j <= n_input; j++)
            if (fscanf(f, "%lf", &weight[i][j]) != 1)
                return FALSE;
    }
    return TRUE;
}

void NNLayer::feed_forward (REAL *input, REAL *output)
{
    assert(input != NULL);
    assert(output != NULL);

    for (size_t i = 0; i < n_neuron; i++) {
        REAL *we = weight[i], *in = input;
        REAL wsum = *we++;
        for (size_t j = n_input; j; j--)
            wsum += (*we++) * (*in++);

        *output++  = sigmoid(wsum);
        sig_der[i] = sigmoid_deriv(wsum);
    }
}

/* will change derin & derout */
void NNLayer::back_propagate (REAL *input, REAL *derin, REAL *derout)
{
    assert(input != NULL);
    assert(derin != NULL);
    assert(derout != NULL);

    rv_clear(derout, n_input);
    for (size_t i = 0; i < n_neuron; i++) {
        REAL di = (*derin++ *= sig_der[i]);
        REAL *de = derout, *we = weight[i];
        REAL *wd = wdelta[i], *in = input;

        *wd++ += di; we++;
        for (size_t j = n_input; j; j--) {
            *de++ += di * (*we++);
            *wd++ += di * (*in++);
        }
    }
}

void NNLayer::update_weight (REAL learn_rate)
{
    size_t i, j;

    for (i = 0; i < n_neuron; i++) {
        for (j = 0; j <= n_input; j++) {
            weight[i][j] -= learn_rate * wdelta[i][j];
            wdelta[i][j] = 0;
        }
    }
}

void NNLayer::initialize (REAL min, REAL max)
{
    assert(min < max);
    winit_min = min;
    winit_max = max;
    initialize();
}

void NNLayer::initialize ()
{
    size_t i, j;

    for (i = 0; i < n_neuron; i++)
        for (j = 0; j <= n_input; j++) {
            weight[i][j] =  winit_min
                + randu() * (winit_max-winit_min);
            wdelta[i][j] = 0;
        }
}

REAL NNLayer::sigmoid (REAL x)
{
    stored_sigmoid = quick_tanh(x);
    return stored_sigmoid;
}

REAL NNLayer::sigmoid_deriv (REAL x)
{
    assert(stored_sigmoid == quick_tanh(x));
    return (1 - stored_sigmoid*stored_sigmoid);
}

/* for line search */
void NNLayer::backup_weight ()
{
    size_t i;
    for (i = 0; i < n_neuron; i++)
        rv_copy(wbackup[i], weight[i], n_input+1);
}

void NNLayer::restore_weight ()
{
    size_t i;
    for (i = 0; i < n_neuron; i++)
        rv_copy(weight[i], wbackup[i], n_input+1);
}

void NNLayer::temp_update_weight (REAL step)
{
    size_t i, j;
    for (i = 0; i < n_neuron; i++)
        for (j = 0; j <= n_input; j++)
            weight[i][j] = wbackup[i][j] - step * wdelta[i][j];
}

/* norm square of gradient */
REAL NNLayer::g_norm2 ()
{
    size_t i, j;
    REAL s = 0;
    for (i = 0; i < n_neuron; i++)
        for (j = 0; j <= n_input; j++)
            s += wdelta[i][j] * wdelta[i][j];
    return s;
}

REAL NNLayer::g_dot_old ()
{
    size_t i, j;
    REAL s = 0;
    for (i = 0; i < n_neuron; i++)
        for (j = 0; j <= n_input; j++)
            s += wdelta[i][j] * wdel_old[i][j];
    return s;
}

void NNLayer::cg_update (REAL beta)
{
    size_t i, j;
    for (i = 0; i < n_neuron; i++)
        for (j = 0; j <= n_input; j++) {
            wdel_old[i][j] = wdelta[i][j];
            wdelta[i][j] += beta * wdir[i][j];
            wdir[i][j] = wdelta[i][j];
        }
}
